/*
 * Copyright (c) 2004 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: main.c,v 1.4 2003/11/22 11:37:29 zinnia Exp $
 */

#include "sps.h"

static void usage()
{
	fprintf(stderr, "usage: i don't know. \n");
	exit(1);
}

/* XXX */
SPS_Context* context;
SPS_Context* SPS_CreateContext(SDL_Surface* s)
{
	SPS_Context* c = malloc(sizeof(SPS_Context));

	if(c == NULL)  return NULL;

	c->lua = lua_open();
	return c;
}

static int sps_mainthread(void* data)
{
	SPS_Context* c = (SPS_Context*)data;

	while(global.state != SPS_STATE_DONE){
		if(global.state == SPS_STATE_PAGELOAD){
			GList* ptr = g_list_nth(context->pages, context->pageno - 1);
			if(ptr == NULL){
				fprintf(stderr, "page %d not found.\n", context->pageno);
			} else {
//				fprintf(stderr, "(%d)>> %s", context->pageno, ptr->data);
				lua_dostring(context->lua, ptr->data);
			}
			SPS_SetState(SPS_STATE_PAGEWAIT);
		}
		SDL_Delay(20);
	}

	return 0;
}

static void read_pages(const char* path)
{
#define READ_BUFFER_SIZE 0x1000

	FILE* fp;
	char* pagebuf; char* ptr;
	int pagebufsize = READ_BUFFER_SIZE;
	int readlen;

	int currentpage = 0;

	char buf[READ_BUFFER_SIZE];
	
	pagebuf = g_malloc(pagebufsize);
	ptr = pagebuf; readlen = 0;

	fp = fopen(path, "r");
	if(fp == NULL) return;

	while(fgets(buf, READ_BUFFER_SIZE - 1, fp)){
		if (!strncmp(buf, "sps_endpage()", 13)){
			/* register */
			char* tmp = g_malloc(readlen + 1);

			memmove(tmp, pagebuf, readlen + 1);
			context->pages = g_list_append(context->pages, tmp);
			ptr = pagebuf; readlen = 0;
		} else {
			int len = strlen(buf);
			if(!strncmp(buf, "sps_beginpage(", 14)){
				currentpage ++;
				context->maxpage = currentpage;
			}
			if(readlen + len >= pagebufsize){
				pagebufsize *= 2;
				pagebuf = g_realloc(pagebuf, pagebufsize);
			}

			memmove(ptr, buf, len + 1);
			ptr += len;
			readlen += len;
		}
	}
	fclose(fp);
	g_free(pagebuf);

	return;
}

static int start(SDL_Surface* screen, const char* path)
{
	SPS_Context* c = SPS_CreateContext(screen);

	if(c == NULL) {
		fprintf(stderr, "malloc failed.\n");
		return -1;
	}

	context = c;
	c->iconv_context = iconv_open("UTF-8", DEFAULT_ENCODING);
	if(c->iconv_context == (iconv_t)-1){
		fprintf(stderr, "iconv_open: %s\n", strerror(errno));
		return -1;
	}

	c->icons = g_hash_table_new(g_str_hash, g_str_equal);
	c->fontpath = g_hash_table_new(g_str_hash, g_str_equal);
	c->font = g_hash_table_new(g_str_hash, g_str_equal);

	c->path = path;

	c->cursor.x = 0;
	c->cursor.y = 0;

	c->in_linerender = 0;
	c->line = NULL;

	lua_context_prepare(c->lua);

	c->surface = screen;
	c->backcolor = SDL_MapRGB(c->surface->format, 255, 255, 255);
	c->forecolor = SDL_MapRGB(c->surface->format, 0, 0, 0);
	c->pages = NULL;

	read_pages(path);

	/* XXX compilation */

	global.mutex = SDL_CreateMutex();
	c->thread = SDL_CreateThread(sps_mainthread, (void*)c);
		
	return 0;
}

/* XXX not called */
static void destroy()
{
	/* XXX release contents */
	
	g_hash_table_destroy(context->icons);
	g_hash_table_destroy(context->fontpath);
	g_hash_table_destroy(context->font);

	SDL_FreeSurface(context->surface);
}

static void parse_event(const SDL_Event* e)
{
	switch(e->type){
	case SDL_QUIT:
		exit(1);
		break;
	case SDL_KEYDOWN:
		if(e->key.keysym.sym == SDLK_ESCAPE){
			exit(1);
		}
		/* key handling */ 
		switch(global.state){
		case SPS_STATE_PAGEWAIT: /* XXX */
			if(context->pageno >= context->maxpage){
				SPS_SetState(SPS_STATE_DONE);
			} else {
				context->pageno ++;
				SPS_SetState(SPS_STATE_PAGELOAD);
			}
			break;
		case SPS_STATE_PAUSE:
			SPS_SetState(SPS_STATE_DRAW);
			break;
		}
		break;
	}
}

SPS_Global global;

static void parse_geometry(const char* expression, int* w, int* h)
{
	sscanf(expression, "%dx%d", w, h);
	return;
}

#define MAXFPS (60)
#define DISPLAY_DELAY (1000 / (MAXFPS))

int main(int argc, char* argv[])
{
	int i;

	char* path = NULL;
	SDL_Surface* work;

	global.fullscreen = 0;
	global.w = 640;
	global.h = 480;
	global.showfps = 0;

	for(i = 1; i < argc; i++){
		switch(*argv[i]){
		case '-':
			switch(*(argv[i] + 1)){
			case 'h':
			case '?':
				usage();
				break;

			case 'g':
				i++;
				parse_geometry(argv[i],
					       &(global.w),
					       &(global.h));
				break;
			case 'f':
				global.fullscreen = 1;
				break;
			case 'F':
				global.showfps = 1;
				break;
			}
			break;
		default:
			path = argv[i];
			break;
		}
	}
	if(path == NULL) usage();

	if(SDL_Init(SDL_INIT_VIDEO) < 0){
		fprintf(stderr, "SDL_Init: %s\n", SDL_GetError());
		return 0;
	}
	if(TTF_Init() < 0){
		fprintf(stderr, "TTF_Init: %s\n", SDL_GetError());
		return 0;
	}

	atexit(SDL_Quit);

	global.display = 
		SDL_SetVideoMode(global.w, 
				 global.h, 
				 32,
				 SDL_SWSURFACE |
				 global.fullscreen ? SDL_FULLSCREEN : 0 );

	if(global.display == NULL){
		fprintf(stderr, "SDL_SetVideoMode: %s\n", SDL_GetError());
		return 0;
	}

	work = SDL_ConvertSurface(global.display, global.display->format, global.display->flags);
	if(work == NULL){
		fprintf(stderr, "SDL_ConvertSurface: %s\n", SDL_GetError());
		return 0;
	}
	if(start(work, path) == -1){
		fprintf(stderr, "%s: initialization failed.\n", path);
	}

	context->pageno = 1;
	SPS_SetState(SPS_STATE_PAGELOAD);

	global.startticks = SDL_GetTicks();
	global.frames = 0;

	while(1){
		SDL_Event e;

		while(SDL_PollEvent(&e)){
			parse_event(&e);
		}
		SDL_Delay(DISPLAY_DELAY);
		switch(global.state){
		case SPS_STATE_DONE:
			goto infinite_loop_end;

		case SPS_STATE_DRAW:
			break;

		case SPS_STATE_PAGEWAIT:
		case SPS_STATE_PAUSE:
			if(global.redraw == 0) break;
			SDL_BlitSurface(context->surface, NULL, global.display, NULL);
			showfps();
			SDL_UpdateRect(global.display, 0, 0, 0, 0);
			global.frames++;
			global.redraw = 0;
			break;
		}
	}
 infinite_loop_end:
	fprintf(stderr, "done.\n");
	return 0;
}
